# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Functions implementing various filters that Capella supports."""
from __future__ import annotations

import collections.abc as cabc
import dataclasses
import importlib
import typing as t
import urllib.parse

from lxml import etree

import capellambse.loader
from capellambse import aird, diagram, model

from .. import _common as c

Phase2CompositeFilter = t.Callable[
    [c.ElementBuilder, diagram.DiagramElement], None
]
CompositeFilter = t.Callable[
    [c.ElementBuilder, diagram.DiagramElement],
    t.Optional[Phase2CompositeFilter],
]
#: Maps composite filter names to phase-1 callables
COMPOSITE_FILTERS: dict[str, CompositeFilter] = {}

GlobalFilter = t.Callable[["FilterArguments", etree._Element], None]
#: Maps names of global filters to functions that implement them
GLOBAL_FILTERS: dict[str, GlobalFilter] = {}

PLUGIN_PATH = (
    "/plugin/org.polarsys.capella.core.sirius.analysis"
    "/description/context.odesign#/"
)

_TDiagramElement = t.TypeVar("_TDiagramElement", bound=diagram.DiagramElement)


def composite_filter(
    name: str,
) -> cabc.Callable[[CompositeFilter], CompositeFilter]:
    """Register a composite filter.

    Composite filters are executed in two phases. The first phase occurs
    during each diagram element's creation. Here the callables
    registered with this decorator are directly executed with the
    diagram under construction and the to-be-filtered element as
    arguments.

    The second phase happens after the entire diagram has been created.
    Filters that want to run in this phase should return another
    callable object during the first phase; this object will then be
    called with the same (now constructed) diagram and the element.

    Composite filters should always operate in place. They may append
    additional elements to the diagram, but they should never remove
    any; instead they should simply change their "hidden" flag to True.
    """

    def add_comp_filter(func: CompositeFilter) -> CompositeFilter:
        COMPOSITE_FILTERS[name] = func
        return func

    return add_comp_filter


def phase2_composite_filter(
    name: str,
) -> cabc.Callable[[Phase2CompositeFilter], Phase2CompositeFilter]:
    """Register a composite filter that only needs to run in phase 2."""

    def add_comp_filter(func: Phase2CompositeFilter) -> Phase2CompositeFilter:
        def phase1dummy(
            _1: c.ElementBuilder,
            _2: diagram.DiagramElement,
        ) -> Phase2CompositeFilter:
            return func

        COMPOSITE_FILTERS[name] = phase1dummy
        return func

    return add_comp_filter


def global_filter(name: str) -> cabc.Callable[[GlobalFilter], GlobalFilter]:
    """Register a global filter."""

    def add_global_filter(func: GlobalFilter) -> GlobalFilter:
        GLOBAL_FILTERS[name] = func
        return func

    return add_global_filter


def setfilters(
    seb: c.SemanticElementBuilder,
    dgobject: _TDiagramElement,
) -> _TDiagramElement:
    """Set the filters on the element.

    This is phase 1 of composite filter execution.

    Parameters
    ----------
    seb
        The element builder instance
    dgobject
        The constructed diagram element on which to set filters.

    Returns
    -------
    dgobject
        The modified ``dgobject``. This is done for convenience; the
        object is modified in place.

    See Also
    --------
    composite_filter : Registering a composite filter.
    """
    dgobject.hidden = seb.diag_element.attrib.get("visible", "true") != "true"
    dgobject.hidelabel = False

    for flt in seb.diag_element.iterchildren("graphicalFilters"):
        flttype = flt.attrib[c.ATT_XMT]
        if flttype == "diagram:HideLabelFilter":
            dgobject.hidelabel = True
        elif flttype in {"diagram:HideFilter", "diagram:CollapseFilter"}:
            dgobject.hidden = True
        elif flttype == "diagram:AppliedCompositeFilters":
            _set_composite_filter(seb, dgobject, flt)
        else:
            c.LOGGER.warning("Ignoring unknown filter type %r", flttype)
    return dgobject


@dataclasses.dataclass(frozen=True)
class FilterArguments:
    """Basic arguments for diagram-filters."""

    target_diagram: diagram.Diagram
    diagram_root: etree._Element
    melodyloader: capellambse.loader.MelodyLoader
    params: dict[str, t.Any]


def applyfilters(args: FilterArguments) -> None:
    """Apply filters on the ``target_diagram``.

    This function performs two tasks.

    Firstly it executes phase 2 of all elements' composite filters; see
    :func:`composite_filter` for more details.

    Secondly it applies the diagram's global filters. These are always
    applied after all composite filters have run.
    """
    # Apply post-processing filters on elements
    for dgobject in args.target_diagram:
        try:
            filters = dgobject._compfilters  # type: ignore[union-attr]
        except AttributeError:
            continue
        assert dgobject.uuid is not None

        flttype: str
        p2flt: Phase2CompositeFilter
        for flttype, p2flt in filters:
            c.LOGGER.debug(
                "Applying post-processing filter %r to %s %r",
                flttype,
                dgobject.styleclass or dgobject.__class__.__name__,
                dgobject.uuid,
            )
            data_element = args.melodyloader[dgobject.uuid]
            p2flt(
                c.ElementBuilder(
                    target_diagram=args.target_diagram,
                    diagram_tree=args.diagram_root,
                    data_element=data_element,
                    melodyloader=args.melodyloader,
                    fragment=args.melodyloader.find_fragment(data_element),
                ),
                dgobject,
            )

        del dgobject._compfilters  # type: ignore[union-attr]

    # Apply global diagram filters
    for flt in args.diagram_root.iterchildren("activatedFilters"):
        try:
            flttype = flt.attrib[c.ATT_XMT]
        except KeyError:
            flttype = "(no filter type given)"
        if flttype != "filter:CompositeFilterDescription":
            c.LOGGER.warning("Unknown global filter type %r", flttype)
            continue

        try:
            fltname = _extract_filter_type(flt)
        except ValueError as err:
            c.LOGGER.debug("Ignoring broken global filter: %s", err)
            continue

        try:
            fltfunc = GLOBAL_FILTERS[fltname]
        except KeyError:
            c.LOGGER.debug("Ignoring unknown global filter %r", fltname)
            continue

        c.LOGGER.debug("Applying global filter %r", fltname)
        fltfunc(args, flt)


def _set_composite_filter(
    seb: c.SemanticElementBuilder,
    dgobject: diagram.DiagramElement,
    flt: etree._Element,
) -> None:
    try:
        flttype = _extract_filter_type(
            next(flt.iterchildren("compositeFilterDescriptions"))
        )
    except (StopIteration, ValueError):
        c.LOGGER.debug(
            "Ignoring unusable composite filter on object %r", dgobject.uuid
        )
        return

    try:
        fltfunc = COMPOSITE_FILTERS[flttype]
    except KeyError:
        c.LOGGER.debug(
            "Ignoring unknown composite filter %r on object %r",
            flttype,
            dgobject.uuid,
        )
        return

    phase2 = fltfunc(seb, dgobject)

    # Register the phase 2 callable
    if phase2 is None:
        return
    if not hasattr(dgobject, "_compfilters"):
        dgobject._compfilters = []  # type: ignore[union-attr]
    dgobject._compfilters.append((flttype, phase2))


def _extract_filter_type(flt_elm: etree._Element) -> str:
    try:
        flttype = flt_elm.attrib["href"]
    except KeyError:
        raise ValueError("Filter element has no href") from None

    compfilter = c.RE_COMPOSITE_FILTER.search(flttype)
    if not compfilter or not compfilter.group(1):
        raise ValueError("Filter href does not match known pattern") from None

    return urllib.parse.unquote(compfilter.group(1))


class ActiveFilters(t.MutableSet[str]):
    """A set of active filters on a Diagram.

    Enable access to set, add and remove active filters on a
    :class:`~capellambse.diagram.Diagram`.
    """

    __slots__ = ("_model", "_target", "_diagram")
    _xml_tag = "activatedFilters"

    def __init__(
        self,
        model: model.MelodyModel,  # pylint: disable=redefined-outer-name
        diagram: model.diagram.Diagram,  # pylint: disable=redefined-outer-name
    ) -> None:
        self._model = model
        self._diagram = diagram
        assert isinstance(diagram._element, etree._Element)
        self._target = self._model._loader[diagram._element.attrib["repPath"]]

    @property
    def _elements(self) -> t.Iterator[etree._Element]:
        return self._target.iterchildren(self._xml_tag)

    @staticmethod
    def _get_filter_name(filter: etree._Element) -> str | None:
        filter_name = c.RE_COMPOSITE_FILTER.search(filter.get("href", ""))
        if filter_name:
            return filter_name.group(1)
        return None

    def __contains__(self, filter: object) -> bool:
        if not isinstance(filter, str):
            return False
        return filter in iter(self)

    def __iter__(self) -> cabc.Iterator[str]:
        for filter in self._elements:
            if filter_name := self._get_filter_name(filter):
                yield filter_name

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def add(self, value: str) -> None:
        """Add an activated filter to the diagram.

        Writes a new ``<activatedFilters>`` XML element to the
        ``<diagram:DSemanticDiagram>`` XML element. If the ``value`` is
        not apparent in :any:`capellambse.aird.GLOBAL_FILTERS` as a key
        it can not be applied when rendering. It should still be visible
        in the GUI.
        """
        if value in self:
            return

        diag_descriptor = aird._build_descriptor(
            self._model._loader, self._diagram._element
        )
        viewpoint = urllib.parse.quote(diag_descriptor.viewpoint)
        assert diag_descriptor.styleclass is not None
        diagclass = urllib.parse.quote(diag_descriptor.styleclass)
        href = "/".join(
            (
                f"platform:{PLUGIN_PATH}",
                f"@ownedViewpoints[name='{viewpoint}']",
                f"@ownedRepresentations[name='{diagclass}']",
                f"@filters[name='{value}']",
            )
        )
        elt = c.ELEMENT.activatedFilters(
            {"href": href, c.ATT_XMT: "filter:CompositeFilterDescription"}
        )
        self._target.append(elt)
        self._diagram.invalidate_cache()

    def discard(self, value: str) -> None:
        """Remove the filter with the given ``value`` from the diagram.

        Deletes ``<activatedFilters>`` XML element from the diagram
        element tree.
        """
        for filter in self._elements:
            filter_name = self._get_filter_name(filter)
            if filter_name is not None and value == filter_name:
                self._target.remove(filter)
                self._diagram.invalidate_cache()
                break

    def __repr__(self) -> str:  # pragma: no cover
        return f"{set(self)!r}"


# Load filter modules
for module in ("composite", "global"):
    try:
        importlib.import_module(f"{__name__}.{module}")
    except Exception as _err:  # pylint: disable=broad-except
        c.LOGGER.error(
            "Cannot load filters from %s: %s: %s",
            module,
            _err.__class__.__name__,
            _err,
        )
del module



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing as t

from svgwrite import container, gradients, path

# pylint: disable-next=unused-import  # compatibility re-export
from capellambse.diagram._icons import _make_lgradient

from . import decorations

Gradient = t.Union[gradients.LinearGradient, gradients.RadialGradient]


def _make_marker(
    ref_pts: tuple[float, float],
    size: tuple[float, float],
    *,
    id_: str,
    d: str,
    **kwargs: t.Any,
) -> container.Marker:
    marker = container.Marker(
        insert=ref_pts,
        size=size,
        id_=id_,
        orient="auto",
        markerUnits="userSpaceOnUse",
    )
    marker.add(path.Path(d=d, **kwargs))
    return marker


@decorations.marker_factories
def fine_arrow_mark(id_, /, **kw) -> container.Marker:
    d = (
        "M 0.4535,0.107 7.309,3.621 0.492,7.407 "
        "0.144,7.407 6.414,3.63 0.136,0.479 Z"
    )
    return _make_marker((7, 3.75), (7.5, 7.5), id_=id_, d=d, **kw)


@decorations.marker_factories
def arrow_mark(id_, /, **kw) -> container.Marker:
    return _make_marker(
        (5, 2.5), (5.5, 5.5), id_=id_, d="M 0,0 5,2.5 0,5", **kw
    )


def _make_diamond_marker(id_: str, /, **kw) -> container.Marker:
    return _make_marker(
        (0, 3), (11, 6), id_=id_, d="M 0,3 5,0.5 10,3 5,5.5 Z", **kw
    )


@decorations.marker_factories
def diamond_mark(id_, /, **kw) -> container.Marker:
    kw.setdefault("fill", "#fff")
    return _make_diamond_marker(id_, **kw)


@decorations.marker_factories
def filled_diamond_mark(id_, /, **kw) -> container.Marker:
    kw.setdefault("fill", "#000")
    return _make_diamond_marker(id_, **kw)


@decorations.marker_factories
def generalization_mark(id_, /, **kw) -> container.Marker:
    kw.setdefault("fill", "#fff")
    d = "M 0.1275,7.5 7.5,3.75 0,0 Z"
    return _make_marker((7, 4), (7.5, 7.5), id_=id_, d=d, **kw)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = [
    "AttributeProperty",
    "BooleanAttributeProperty",
    "DatetimeAttributeProperty",
    "EnumAttributeProperty",
    "HTMLAttributeProperty",
    "NumericAttributeProperty",
]
import collections.abc as cabc
import datetime
import enum
import math
import os
import re
import sys
import typing as t

import markupsafe

from capellambse import helpers


class AttributeProperty:
    """A property that forwards access to the underlying XML element."""

    __slots__ = (
        "attribute",
        "default",
        "writable",
        "returntype",
        "__name__",
        "__objclass__",
        "__dict__",
    )

    NOT_OPTIONAL = object()

    def __init__(
        self,
        attribute: str,
        *,
        returntype: cabc.Callable[[str], t.Any] = str,
        optional: bool = False,
        default: t.Any = None,
        writable: bool = True,
        __doc__: str | None = None,
    ) -> None:
        """Create an AttributeProperty.

        Parameters
        ----------
        attribute
            The attribute on the XML element to handle.
        returntype
            The type to return the result as. Must accept a single
            ``str`` as argument.
        optional
            If False (default) and the XML attribute does not exist, an
            AttributeError is raised. Otherwise a default value is
            returned.
        default
            A new-style format string to use as fallback value. You can
            access the object instance as ``self`` and the XML element
            as ``xml``.
        writable
            Whether to allow modifying the XML attribute.
        """
        self.attribute = attribute
        self.writable = writable
        self.returntype = returntype
        if optional or default is not None:
            self.default = default
        else:
            self.default = self.NOT_OPTIONAL

        self._original_value: t.Any = None

        self.__name__ = "(unknown)"
        self.__objclass__: type[t.Any] | None = None
        self.__doc__ = __doc__

    @property
    def _qualname(self) -> str:
        """Generate the qualified name of this descriptor."""
        if self.__objclass__ is None:
            return f"(unknown {type(self).__name__} - call __set_name__)"
        return f"{self.__objclass__.__name__}.{self.__name__}"

    @t.overload
    def __get__(self, obj: None, objtype: type) -> AttributeProperty: ...

    @t.overload
    def __get__(self, obj: t.Any, objtype: type | None = None) -> t.Any: ...

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        rv = self._get(obj)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def _get(self, obj):
        try:
            rv = self.returntype(obj._element.attrib[self.attribute])
        except KeyError:
            if self.default is not self.NOT_OPTIONAL:
                return self.default and self.returntype(
                    self.default.format(self=obj, xml=obj._element)
                )
            raise TypeError(
                f"Mandatory XML attribute {self.attribute!r} not found on"
                f" {obj._element!r}"
            ) from None
        return rv

    def __set__(self, obj, value) -> None:
        stringified = self._stringify(obj, value)
        if obj._constructed:
            sys.audit(
                "capellambse.setattr",
                obj,
                self.__name__,
                self._original_value or value,
            )
        if stringified is None:
            self.__delete__(obj)
        else:
            obj._element.attrib[self.attribute] = stringified

    def _stringify(self, obj, value) -> str | None:
        if not self.writable and obj._element.get(self.attribute) is not None:
            raise TypeError(
                f"Cannot set attribute {self.__name__!r} on"
                f" {type(obj).__name__!r} objects"
            )

        if value == self.default:
            return None

        stringified = str(value)

        try:
            roundtripped = self.returntype(stringified)
        except (TypeError, ValueError) as err:
            raise TypeError(
                "Value is not round-trip safe:"
                f" Cannot read back {stringified} as {value}"
            ) from err

        if roundtripped != value:
            raise TypeError(
                "Value is not round-trip safe:"
                f" {value!r} would be read back as {roundtripped!r}"
            )

        return stringified

    def __delete__(self, obj: t.Any) -> None:
        if not self.writable:
            raise TypeError(
                f"Cannot delete attribute {self.__name__!r} on"
                f" {type(obj).__name__!r} objects"
            )

        try:
            del obj._element.attrib[self.attribute]
        except KeyError:
            pass

    def __set_name__(self, owner: type[t.Any], name: str) -> None:
        self.__name__ = name
        self.__objclass__ = owner


class HTMLAttributeProperty(AttributeProperty):
    """An AttributeProperty that gracefully handles HTML-in-XML."""

    def __init__(
        self,
        attribute: str,
        *,
        optional: bool = False,
        writable: bool = True,
        __doc__: str | None = None,
    ) -> None:
        super().__init__(
            attribute,
            default="",
            returntype=markupsafe.Markup,
            optional=optional,
            writable=writable,
            __doc__=__doc__,
        )

    @t.overload
    def __get__(self, obj: None, objtype: type) -> AttributeProperty: ...
    @t.overload
    def __get__(self, obj: t.Any, objtype: type | None = None) -> t.Any: ...
    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:
            return self

        rv = self._get(obj)

        if os.getenv("CAPELLAMBSE_XHTML") == "1":
            rv = helpers.repair_html(rv)

        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(self, obj: t.Any, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(
                f"Cannot set {self._qualname} on {obj!r}:"
                f" Expected str, got {type(value).__name__}"
            )
        value = helpers.repair_html(value)
        super().__set__(obj, value)


class NumericAttributeProperty(AttributeProperty):
    """Attribute property that handles (possibly infinite) numeric values.

    Positive infinity is stored in Capella XML as `*`. This class takes
    care of converting to and from that value when setting or retrieving
    the value.

    Note that there is currently no representation of negative infinity,
    which is why ``-inf`` is rejected with a :class:`ValueError`.

    ``NaN`` values are rejected with a ValueError as well.
    """

    def __init__(
        self,
        attribute: str,
        *,
        optional: bool = False,
        default: int | float | None = None,
        allow_float: bool = True,
        writable: bool = True,
        __doc__: str | None = None,
    ) -> None:
        super().__init__(
            attribute,
            optional=optional,
            default=default,
            writable=writable,
            __doc__=__doc__,
        )
        self.number_type = float if allow_float else int

        self._original_value: None | float | int = None

    def __get__(self, obj, objtype=None):
        value = super().__get__(obj, objtype)
        if not isinstance(value, str):
            return value

        if value == "*":
            return math.inf
        return self.number_type(value)

    def __set__(self, obj, value) -> None:
        if not isinstance(value, (int, float)):
            raise TypeError(f"Not a number: {value}")

        if value == math.inf:
            strvalue = "*"
        elif value == -math.inf:
            raise ValueError("Cannot set value to negative infinity")
        else:
            strvalue = str(self.number_type(value))

        self._original_value = value
        super().__set__(obj, strvalue)


class BooleanAttributeProperty(AttributeProperty):
    """An AttributeProperty that works with booleans."""

    def __init__(
        self,
        attribute: str,
        *,
        writable: bool = True,
        __doc__: str | None = None,
    ) -> None:
        super().__init__(
            attribute,
            returntype=lambda x: x,
            optional=False,
            writable=writable,
            __doc__=__doc__,
        )

        self._original_value: None | bool = None

    @t.overload
    def __get__(self, obj: None, objtype: type) -> AttributeProperty: ...

    @t.overload
    def __get__(self, obj: t.Any, objtype: type | None = None) -> bool: ...

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        rv = obj._element.get(self.attribute, "false") == "true"
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv

    def __set__(self, obj, value) -> None:
        self._original_value = value
        if value:
            super().__set__(obj, "true")
        else:
            self.__delete__(obj)


class DatetimeAttributeProperty(AttributeProperty):
    """An AttributeProperty that stores a datetime.

    The value stored in the XML will be formatted as required by
    Capella. This format is the ISO8601 format with millisecond
    precision, but no ``:`` in the time zone specification.
    """

    __slots__ = ("format",)

    re_set = re.compile(r"(?<=[+-]\d\d):(?=\d\d$)")
    re_get = re.compile(r"(?<=[+-]\d\d)(?=\d\d$)")

    def __init__(
        self,
        attribute: str,
        *,
        optional: bool = True,
        writable: bool = True,
        __doc__: str | None = None,
    ) -> None:
        super().__init__(
            attribute,
            optional=optional,
            writable=writable,
            __doc__=__doc__,
        )

        self._original_value: None | datetime.datetime = None

    @t.overload
    def __get__(
        self, obj: None, objtype: type
    ) -> DatetimeAttributeProperty: ...

    @t.overload
    def __get__(
        self, obj: t.Any, objtype: type | None = None
    ) -> datetime.datetime | None: ...

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        formatted = super().__get__(obj, objtype)
        if formatted is None:
            return None
        formatted = self.re_get.sub(":", formatted)
        return datetime.datetime.fromisoformat(formatted)

    def __set__(self, obj, value) -> None:
        if value is None:
            self.__delete__(obj)
        elif isinstance(value, datetime.datetime):
            self._original_value = value
            if value.tzinfo is None or value.tzinfo.utcoffset(value) is None:
                value = value.astimezone()

            formatted = value.isoformat("T", "milliseconds")
            formatted = self.re_set.sub("", formatted)
            super().__set__(obj, formatted)
        else:
            raise TypeError(f"Expected datetime, not {type(value).__name__}")


class EnumAttributeProperty(AttributeProperty):
    """An AttributeProperty whose values are determined by an Enum.

    This works in much the same way as the standard AttributeProperty,
    except that the returned and consumed values are not simple strings,
    but members of the Enum that was passed into the constructor.

    Usually it is expected that the enum members will be directly
    assigned to this property. However it is also possible to assign a
    :class:`str` instead. In this case, the string will be taken to be
    an enum member's name. In both cases, the enum member's value will
    be placed in the underlying XML attribute.

    If the XML attribute contains a value that does not correspond to
    any of the Enum's members, a KeyError will be raised. If the
    attribute is completely missing from the XML and there was no
    ``default=`` value set during construction, this property will
    return ``None``.
    """

    __slots__ = ("enumcls",)

    def __init__(
        self,
        attribute: str,
        enumcls: type[enum.Enum],
        *args: t.Any,
        default: str | enum.Enum | None = None,
        **kw: t.Any,
    ) -> None:
        """Create an EnumAttributeProperty.

        Parameters
        ----------
        attribute
            The attribute on the XML element to handle.
        enumcls
            The :class:`enum.Enum` subclass to use. The class' members'
            values are used as the possible values for the XML
            attribute.
        default
            The default value to return if the attribute is not present
            in the XML. If None, an AttributeError will be raised
            instead.
        """
        if not (isinstance(enumcls, type) and issubclass(enumcls, enum.Enum)):
            raise TypeError(
                f"enumcls must be an Enum subclass, not {enumcls!r}"
            )

        if default is None or isinstance(default, enumcls):
            pass
        elif isinstance(default, str):
            default = enumcls[default]
        else:
            raise TypeError(
                f"default must be a member (or its name) of {enumcls!r},"
                f" not {default!r}"
            )
        super().__init__(
            attribute, *args, optional=True, default=default, **kw
        )
        self.enumcls = enumcls

    def __get__(self, obj: t.Any, objtype: type | None = None) -> t.Any:
        if obj is None:
            return self

        try:
            rawvalue = super().__get__(obj, objtype)
        except AttributeError:
            if self.default is not None:
                return self.default
            raise

        if rawvalue is None:
            return None
        return self.enumcls(rawvalue)

    def __set__(self, obj: t.Any, value: str | enum.Enum) -> None:
        assert self.__objclass__ is not None
        if isinstance(value, str):
            try:
                value = self.enumcls[value]
            except KeyError:
                raise ValueError(
                    f"{value!r} is not a valid value for {self._qualname}"
                ) from None
        elif not isinstance(value, self.enumcls):
            raise TypeError(
                f"Expected str or member of {self.enumcls}, not {value!r}"
            )

        assert isinstance(value.value, str)
        return super().__set__(obj, value.value)



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Functions for parsing and interacting with diagrams in a Capella model."""
from __future__ import annotations

__all__ = [
    "DiagramDescriptor",
    "DRepresentationDescriptor",
    "ActiveFilters",
    "enumerate_descriptors",
    "enumerate_diagrams",
    "find_target",
    "get_styleclass",
    "parse_diagrams",
    "parse_diagram",
    "iter_visible",
    "GLOBAL_FILTERS",
    "GlobalFilter",
]

import collections.abc as cabc
import logging
import pathlib
import typing as t
import urllib.parse

from lxml import etree

import capellambse._namespaces as _n
from capellambse import diagram, helpers, loader

from . import _common as C
from . import _filters, _semantic, _visual
from ._filters import GLOBAL_FILTERS, ActiveFilters, GlobalFilter

DRepresentationDescriptor = t.NewType(
    "DRepresentationDescriptor", etree._Element
)
r"""A representation descriptor.

These are specific :class:`~lxml.etree._Element`\ s found in AIRD files,
which contain metadata about diagrams.
"""

DIAGRAM_ROOTS = {
    f"{{{_n.NAMESPACES['sequence']}}}SequenceDDiagram",
    f"{{{_n.NAMESPACES['diagram']}}}DSemanticDiagram",
}
"""Representations whose root element has one of these tags are diagrams.

Other representations (e.g. data tables) will not be listed by
`enumerate_diagrams`.
"""


class DiagramDescriptor(t.NamedTuple):
    fragment: pathlib.PurePosixPath
    name: str
    styleclass: str | None
    descriptor: etree._Element
    uid: str
    viewpoint: str
    target: etree._Element


def enumerate_descriptors(
    model: loader.MelodyLoader,
    *,
    viewpoint: str | None = None,
) -> cabc.Iterator[DRepresentationDescriptor]:
    """Enumerate the representation descriptors in the model.

    Parameters
    ----------
    model
        The MelodyLoader instance
    viewpoint
        Only return diagrams of the given viewpoint. If not given, all
        diagrams are returned.
    """
    for view in model.xpath(C.XP_VIEWS):
        if viewpoint and _viewpoint_of(view) != viewpoint:
            continue

        for d in view.iterchildren("ownedRepresentationDescriptors"):
            if d.get(helpers.ATT_XMT) != "viewpoint:DRepresentationDescriptor":
                continue
            rep_path = d.attrib["repPath"]
            if not rep_path.startswith("#"):
                raise RuntimeError(
                    f"Malformed diagram reference: {rep_path!r}"
                )
            diag_root = model[rep_path]
            if diag_root.tag not in DIAGRAM_ROOTS:
                continue

            yield t.cast(DRepresentationDescriptor, d)


def viewpoint_of(descriptor: DRepresentationDescriptor) -> str:
    assert isinstance(descriptor, etree._Element)
    view = descriptor.getparent()
    if view is None:
        raise RuntimeError("No parent view found for diagram")
    return _viewpoint_of(view)


def _viewpoint_of(view: etree._Element) -> str:
    viewpoint = helpers.xpath_fetch_unique(
        "./viewpoint", view, "viewpoint description"
    )
    viewname = C.RE_VIEWPOINT.search(viewpoint.attrib["href"])
    if not viewname:
        return ""
    return urllib.parse.unquote(viewname.group(1))


def enumerate_diagrams(
    model: loader.MelodyLoader,
) -> cabc.Iterator[DiagramDescriptor]:
    """Enumerate the diagrams in the model.

    Parameters
    ----------
    model
        The MelodyLoader instance
    """
    import warnings

    warnings.warn(
        "enumerate_diagrams is deprecated, use enumerate_descriptors instead",
        DeprecationWarning,
        stacklevel=2,
    )

    for descriptor in enumerate_descriptors(model):
        try:
            d = _build_descriptor(model, descriptor)
        except Exception as err:
            C.LOGGER.warning(
                "Ignoring invalid diagram %r: %s", descriptor, err
            )
        else:
            yield d


def parse_diagrams(
    model: loader.MelodyLoader, **params: t.Any
) -> cabc.Iterator[diagram.Diagram]:
    """Parse all diagrams from the model."""
    for descriptor in enumerate_descriptors(model):
        try:
            d = _build_descriptor(model, descriptor)
        except Exception as err:
            C.LOGGER.warning(
                "Ignoring invalid diagram %r: %s", descriptor, err
            )
            continue
        yield parse_diagram(model, d, **params)


def _build_descriptor(
    model: loader.MelodyLoader,
    descriptor: DRepresentationDescriptor,
) -> DiagramDescriptor:
    assert isinstance(descriptor, etree._Element)

    diag_root = model[descriptor.attrib["repPath"]]
    styleclass = get_styleclass(descriptor)
    target = find_target(model, descriptor)

    return DiagramDescriptor(
        fragment=model.find_fragment(descriptor),
        name=descriptor.attrib["name"],
        styleclass=styleclass,
        uid=diag_root.attrib["uid"],
        viewpoint=viewpoint_of(descriptor),
        target=target,
        descriptor=descriptor,
    )


def find_target(
    model: loader.MelodyLoader, descriptor: DRepresentationDescriptor
) -> etree._Element:
    assert isinstance(descriptor, etree._Element)
    target_anchors = list(descriptor.iterchildren("target"))
    if len(target_anchors) != 1:
        raise RuntimeError(
            f"Expected 1 <target> anchor, found {len(target_anchors)}"
        )
    target_href = target_anchors[0].get("href")
    if not target_href:
        raise RuntimeError("<target> anchor has no href")
    return model.follow_link(descriptor[1], target_href)


def get_styleclass(descriptor: DRepresentationDescriptor) -> str | None:
    assert isinstance(descriptor, etree._Element)
    styledescription = helpers.xpath_fetch_unique(
        "./description[@href]", descriptor, "style description"
    )
    # This style description is something resembling XPath, e.g.
    #     platform:/[...]#//[...]/@ownedRepresentations[name='$$$']
    # The thing we're interested in is denoted as $$$ above.
    styleclass_match = C.RE_STYLECLASS.search(styledescription.attrib["href"])
    if not styleclass_match:
        return None
    return urllib.parse.unquote(styleclass_match.group(1))


def parse_diagram(
    model: loader.MelodyLoader,
    descriptor: DRepresentationDescriptor | DiagramDescriptor,
    **params: t.Any,
) -> diagram.Diagram:
    """Parse a single diagram from the model.

    Parameters
    ----------
    model
        A loaded model.
    descriptor
        A DiagramDescriptor as obtained from :func:`enumerate_diagrams`.
    """
    if not isinstance(descriptor, DiagramDescriptor):
        descriptor = _build_descriptor(model, descriptor)

    diag = diagram.Diagram(
        descriptor.name, styleclass=descriptor.styleclass, uuid=descriptor.uid
    )
    dgtree = model.follow_link(
        model.trees[descriptor.fragment].root, descriptor.uid
    )
    treedata = helpers.xpath_fetch_unique(
        C.XP_ANNOTATION_ENTRIES,
        dgtree,
        "ownedAnnotationsEntries with source GMF_DIAGRAMS",
        dgtree.attrib["uid"],
    )
    assert treedata is not None

    if treedata.attrib.get("measurementUnit", "Pixel") != "Pixel":
        C.LOGGER.warning(
            "Unknown measurement unit %r",
            treedata.attrib.get("measurementUnit"),
        )

    for data_elm in treedata.iterdescendants("children", "edges"):
        try:
            elm = _element_from_xml(
                C.ElementBuilder(
                    target_diagram=diag,
                    diagram_tree=dgtree,
                    data_element=data_elm,
                    melodyloader=model,
                    fragment=descriptor.fragment,
                )
            )
        except C.SkipObject:
            continue
        assert elm is not None
        diag.add_element(elm, False, force=True)

    if len(diag) == 0:
        C.LOGGER.error(
            "Deserialized diagram %r is empty", descriptor.name or diag.name
        )
    else:
        diag.calculate_viewport()
        _filters.applyfilters(
            _filters.FilterArguments(diag, dgtree, model, params)
        )
    return diag


def _element_from_xml(ebd: C.ElementBuilder) -> diagram.DiagramElement:
    """Construct a single diagram element from the model XML."""
    element = ebd.data_element.get("element")
    tag = ebd.melodyloader[element].tag if element else None
    if element is not None and tag != "ownedRepresentationDescriptors":
        factory = _semantic.from_xml
    else:
        factory = _visual.from_xml
    return factory(ebd)


def iter_visible(
    model: loader.MelodyLoader,
    descriptor: DRepresentationDescriptor | DiagramDescriptor,
) -> cabc.Iterator[etree._Element]:
    r"""Iterate over all semantic elements that are visible in a diagram.

    This is a much faster alternative to calling :func:`parse_diagram`
    and iterating over the diagram elements, if you only need to know
    which semantic elements are visible, but are not otherwise
    interested in the layout of the diagram.

    Parameters
    ----------
    model
        A loaded model.
    descriptor
        A DiagramDescriptor as obtained from :func:`enumerate_diagrams`.

    Raises
    ------
    ValueError
        If the corresponding data or style element can't be found in the
        ``*.aird`` file.

    Yields
    ------
    etree._Element
        A semantic element from the ``*.capella`` file.
    """
    if isinstance(descriptor, DiagramDescriptor):
        diag_element = model.follow_link(
            model.trees[descriptor.fragment].root, descriptor.uid
        )
    else:
        assert isinstance(descriptor, etree._Element)
        diag_element = model.follow_link(
            descriptor, descriptor.attrib["repPath"]
        )

    style_data = helpers.xpath_fetch_unique(
        C.XP_ANNOTATION_ENTRIES,
        diag_element,
        "ownedAnnotationsEntries with source GMF_DIAGRAMS",
        diag_element.attrib["uid"],
    )
    port_tag = "ownedBorderedNodes"
    visited: set[str] = set()
    for elt in diag_element.iterdescendants("ownedDiagramElements", port_tag):
        style_element = helpers.xpath_fetch_unique(
            f".//*[@element='{elt.attrib['uid']}']",
            style_data,
            "style description",
        )
        if style_element.get("visible", "true") != "true":
            continue

        # Component Ports have their visible attribute on a child element
        if elt.tag == port_tag:
            try:
                real_se = next(style_element.iterchildren("children"))
            except StopIteration:
                pass

            if real_se.get("visible", "true") != "true":
                continue

        try:
            target = next(elt.iterdescendants("target"))
        except StopIteration:
            C.LOGGER.warning(
                "No semantic element found for %r",
                elt.get("name", elt.attrib["uid"]),
            )
            continue

        try:
            elem = model.follow_link(target, target.attrib["href"])
        except KeyError:
            C.LOGGER.debug(
                "Semantic element has been deleted, ignoring: %s",
                target.attrib["href"],
            )
            continue

        fragment = model.find_fragment(elem)
        if (
            model.trees[fragment].fragment_type == loader.FragmentType.SEMANTIC
            and elem.attrib["id"] not in visited
        ):
            yield elem
            visited.add(elem.attrib["id"])


def is_representation_descriptor(
    e: etree._Element, /
) -> t.TypeGuard[DRepresentationDescriptor]:
    return helpers.xtype_of(e) == "viewpoint:DRepresentationDescriptor"


if not t.TYPE_CHECKING:

    def __getattr__(key: str) -> t.Any:
        if key == "diagram":
            obj = diagram
        else:
            try:
                obj = getattr(diagram, key)
            except NameError:
                try:
                    obj = importlib.import_module(f"capellambse.diagram.{key}")
                except ImportError:
                    raise NameError(f"No name {key} in {__name__}") from None
        import warnings

        warnings.warn(
            (
                f"{__name__}.{key} is deprecated,"
                " import it from capellambse.diagram instead"
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        return obj



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Validation functions for PVMT."""

import logging
import operator
import re

import capellambse._namespaces as _n
from capellambse import helpers

from . import exceptions

LOGGER = logging.getLogger(__name__)

SCOPE_RE = re.compile(r"^\[([A-Z]+)\](.*)\[/\1\]$")
SCOPE_ARCH_MAP = {
    "OPERATIONAL": "org.polarsys.capella.core.data.oa:OperationalAnalysis",
    "SYSTEM": "org.polarsys.capella.core.data.ctx:SystemAnalysis",
    "LOGICAL": "org.polarsys.capella.core.data.la:LogicalArchitecture",
    "PHYSICAL": "org.polarsys.capella.core.data.pa:PhysicalArchitecture",
    "EPBS": "org.polarsys.capella.core.data.epbs:EPBSArchitecture",
}
SCOPE_CLASS_RE = re.compile(r"^(.*)/([^/]+)$")
SCOPE_PROP_OPERATORS = {
    "=": operator.eq,
    "!": operator.ne,
    "?": operator.contains,
    "&": lambda a, b: a.startswith(b),
    "#": lambda a, b: a.endswith(b),
    "<": operator.lt,
    ">": operator.gt,
    "%": operator.le,
    ":": operator.ge,
}
SCOPE_PROP_RE = re.compile(
    r"^(?P<group>[^.]+\.[^.]+)\.(?P<prop>[^.]+)"
    f'(?P<op>[{"".join(SCOPE_PROP_OPERATORS)}])'
    "(?P<val>.*)"
)


def validate_group_scope(pvmt_ext, groupdef, xml_element):
    """Verify that the ``groupdef``'s scope applies to the given element."""
    for scopeline in groupdef.scope.splitlines():
        scope = SCOPE_RE.match(scopeline)
        if scope is None:
            LOGGER.warning("Malformed scope description: %r", scopeline)
            continue

        try:
            func = VALIDATION_FUNCTIONS[scope.group(1)]
        except KeyError:
            LOGGER.warning("Unknown scope tag %r", scope.group(1))
            continue

        try:
            inscope = func(pvmt_ext, scope.group(2), xml_element)
        except AssertionError:
            LOGGER.warning(
                "Failed to parse [%s] scope %r", scope.group(1), scope.group(2)
            )
            continue

        if not inscope:
            raise exceptions.ScopeError(
                f"Element {xml_element!r} is out of scope for PV group "
                f"{groupdef.parent.name}.{groupdef.name}"
            )


def _validate_group_scope_arch(pvmt_ext, scopedesc, xml_element):
    arch_candidates = {
        helpers.xtype_of(i) for i in pvmt_ext.model.iterancestors(xml_element)
    }
    return bool(
        arch_candidates & {SCOPE_ARCH_MAP[a] for a in scopedesc.split(";")}
    )


def _validate_group_scope_class(pvmt_ext, scopedesc, xml_element):
    del pvmt_ext
    match = SCOPE_CLASS_RE.match(scopedesc)
    assert match is not None

    nskey = _n.get_namespace_prefix(match.group(1))
    return (
        xml_element.get(f"{{{_n.NAMESPACES['xsi']}}}type", "")
        == f"{nskey}:{match.group(2)}"
    )


def _validate_group_scope_prop(pvmt_ext, scopedesc, xml_element):
    match = SCOPE_PROP_RE.match(scopedesc)
    assert match is not None

    try:
        prop = pvmt_ext.get_element_pv(
            xml_element, match.group("group"), create=False
        )
        proptype = prop.get_definition(match.group("prop"))
        propval = prop[match.group("prop")]
    except (KeyError, ValueError):
        return False

    try:
        compare = SCOPE_PROP_OPERATORS[match.group("op")]
    except KeyError:
        raise NotImplementedError(
            f"Comparison operator {proptype!r} is not implemented"
        ) from None

    return compare(
        proptype.serialize(propval), proptype.serialize(match.group("val"))
    )


VALIDATION_FUNCTIONS = {
    "ARCHITECTURE": _validate_group_scope_arch,
    "CLASS": _validate_group_scope_class,
    "PROPERTY": _validate_group_scope_prop,
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Functions for receiving the styleclass from a given model object."""
from __future__ import annotations

__all__ = ["get_styleclass"]

import collections.abc as cabc

from capellambse import model


def get_styleclass(obj: model.ModelObject) -> str:
    """Return the styleclass for an individual model object.

    Parameters
    ----------
    obj
        An object received from querying the High-Level API.

    Returns
    -------
    str
        A string used for styling and decorating given ``obj`` in a
        diagram representation.
    """
    styleclass_factory = _STYLECLASSES.get(type(obj).__name__, _default)
    return styleclass_factory(obj)


def _default(obj: model.ModelObject) -> str:
    if isinstance(obj, model.GenericElement):
        return obj.xtype.split(":")[-1]
    return type(obj).__name__


def _association(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.information.Association)
    default_kind = kind = "ASSOCIATION"
    assert isinstance(obj.roles, model.ElementList)
    for member in obj.roles:
        if member.kind != default_kind:
            kind = member.kind.name
    return kind.capitalize()


def _functional_chain_involvement(obj: model.ModelObject) -> str:
    assert isinstance(
        obj,
        (
            model.fa.FunctionalChainInvolvementLink,
            model.fa.FunctionalChainInvolvementFunction,
        ),
    )
    styleclass = _default(obj)
    if isinstance(obj.parent, model.oa.OperationalProcess):
        return styleclass.replace("Functional", "Operational")
    return styleclass


def _functional_exchange(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.fa.FunctionalExchange)
    styleclass = _default(obj)
    if get_styleclass(obj.target) == "OperationalActivity":
        return styleclass.replace("Functional", "Operational")
    return styleclass


def _generic_component(obj: model.ModelObject, extra: str = "") -> str:
    assert isinstance(obj, model.cs.Component)
    styleclass = _default(obj)
    return "".join(
        (
            styleclass[: -len("Component")],
            "Human" * obj.is_human,
            extra,
            ("Component", "Actor")[obj.is_actor],
        )
    )


def _physical_component(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.pa.PhysicalComponent)
    nature = (obj.nature.name, "")[obj.nature.name == "UNSET"]
    return _generic_component(obj, extra=nature.capitalize())


def _part(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.cs.Part)
    assert not isinstance(obj.type, model.ElementList)
    xclass = _default(obj.type)
    if xclass == "PhysicalComponent":
        return _physical_component(obj)
    elif xclass == "Entity":
        return xclass
    return _generic_component(obj.type)


def _port_allocation(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.information.PortAllocation)
    styleclasses = {
        get_styleclass(p)
        for p in (obj.source, obj.target)
        if not isinstance(p, (model.fa.ComponentPort, model.ElementList))
    }
    return f"{'_'.join(sorted(styleclasses))}Allocation"


_STYLECLASSES: dict[str, cabc.Callable[..., str]] = {
    "Association": _association,
    "CapellaIncomingRelation": lambda _: "RequirementRelation",
    "CapellaOutgoingRelation": lambda _: "RequirementRelation",
    "Class": lambda o: "Primitive" * o.is_primitive + "Class",
    "ComponentPort": lambda o: f"CP_{o.direction or 'UNSET'}",
    "ControlNode": lambda o: o.kind.name.capitalize() + _default(o),
    "Entity": lambda o: (
        ("Entity", "OperationalActor")[o.is_actor and o.is_human]
    ),
    "FunctionalChainInvolvementFunction": _functional_chain_involvement,
    "FunctionalChainInvolvementLink": _functional_chain_involvement,
    "FunctionalExchange": _functional_exchange,
    "FunctionInputPort": lambda _: "FIP",
    "FunctionOutputPort": lambda _: "FOP",
    "LogicalComponent": _generic_component,
    "Part": _part,
    "PhysicalComponent": _physical_component,
    "PhysicalPort": lambda _: "PP",
    "Property": lambda o: o.kind.name.capitalize(),
    "PortAllocation": _port_allocation,
    "Region": lambda o: _default(o.parent) + _default(o),
    "SystemComponent": _generic_component,
}



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

import json
import pathlib
import sys

import pytest

import capellambse
from capellambse import aird, diagram, loader

EMPTY_MODEL = pathlib.Path(__file__).parent.joinpath(
    "data/decl/empty_project_52/empty_project_52.aird"
)


class TestAIRDBasicFunctionality:
    test_model = (
        pathlib.Path(__file__).parent / "data" / "parser" / "TestItems.aird"
    )
    test_diagram = "[LAB] Logical System"
    test_json = test_model.with_suffix(".json")
    test_repr = test_model.with_suffix(".repr.txt")
    test_txt = test_model.with_suffix(".txt")

    @pytest.fixture
    def model(self):
        return loader.MelodyLoader(self.test_model)

    @pytest.fixture
    def diagram_under_test(self, model):
        descriptor = next(aird.enumerate_diagrams(model), None)
        assert descriptor is not None
        yield aird.parse_diagram(model, descriptor)

    def test_parsing_all_diagrams_does_not_raise_exceptions(
        self, model, caplog
    ):
        del caplog
        i = 0
        for i, _ in enumerate(aird.parse_diagrams(model), start=1):
            pass
        assert i == 1

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_json_output_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_json.read_text().strip()
        encoder = diagram.DiagramJSONEncoder(indent=4)

        actual = encoder.encode(diagram_under_test)

        assert actual == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_plain_text_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_txt.read_text()
        actual = str(diagram_under_test)
        assert actual + "\n" == expected

    @pytest.mark.xfail(
        sys.platform not in {"win32", "cygwin"},
        reason="Expected rendering inaccuracies on non-Windows platforms",
    )
    @pytest.mark.skip(reason="Currently broken")
    def test_python_code_representation_matches_expected_output(
        self, diagram_under_test, caplog
    ):
        del caplog
        expected = self.test_repr.read_text().strip()
        actual = repr(diagram_under_test)
        assert actual == expected

    def test_enumerate_diagrams_doesnt_crash_if_there_are_no_diagrams(self):
        model = loader.MelodyLoader(EMPTY_MODEL)
        for _ in aird.enumerate_diagrams(model):
            pass


def test_airdparser_msm_produces_valid_json_without_error(
    model: capellambse.MelodyModel,
):
    diag_name = "[MSM] States of Functional Human Being"
    all_diagrams = aird.enumerate_diagrams(model._loader)
    descriptor = next(i for i in all_diagrams if i.name == diag_name)
    parsed = aird.parse_diagram(model._loader, descriptor)

    generated_json = diagram.DiagramJSONEncoder(indent=4).encode(parsed)
    json.loads(generated_json)


@pytest.mark.parametrize(
    ["diag_uid", "num_nodes"],
    [
        ("_7Ft7QKrxEeqOgqWuHJrXFA", 34),
        ("_KLGoEKyJEeqCdMaqCWkrKg", 14),
        ("_be95kNYrEeqiU8uzTY0Puw", 11),
        ("_9i56AagWEeujco-rU7ZOtA", 1),
        ("_GpCVADaWEey5m4gkio8hew", 2),
        ("_-NU3gV2mEey8erljvkS-pQ", 13),
        ("_KZRQIWA9Eey8erljvkS-pQ", 15),
        ("_jofb8WDdEey8erljvkS-pQ", 5),
        ("_bpuRwWDhEey8erljvkS-pQ", 4),
        ("_OAjYgXPIEeyW3OIB4qRWZA", 18),
        ("_RSYVoXPIEeyW3OIB4qRWZA", 8),
        ("_bNNAAXPIEeyW3OIB4qRWZA", 6),
        ("_sCBdcMn-EeyS2Zr7ZWFrXA", 7),
        ("_1aihhGpbEe2Scdb2k-Fvhg", 17),
        ("_fr69QBKnEeuBCogvtwwNBw", 42),
        ("_erzycVIqEeyiRNlyKPJwqw", 44),
        ("_8BQYEYQzEeyxv9w6U6_UQg", 6),
        ("_amKHoKg0EeujRPrkuugYGw", 5),
        ("_1sGtYOSREeulYvEGRaHazg", 5),
        ("_APOQ0QPhEeynfbzU12yy7w", 11),
        ("_K-BhAIKBEeyd4IW8P6K-AQ", 14),
        ("_0hjtgB4cEe2Jgqeci1tAhA", 1),
        ("_hJD6gaWXEeqh-bahrg4jkA", 18),
        ("_u8_YMLB-EeqP7JXhmLvOsg", 26),
        ("_2TLuoAL8EeuogvvqBpYPnQ", 6),
        ("_KVGPoKJwEeusg-blS2y7tg", 18),
        ("_ZN1JAKJxEeusg-blS2y7tg", 21),
        ("_mKZeQKZwEeuRO5WMG4dh5w", 2),
        ("_MDzzEAWrEeyIAukYgdUX5A", 11),
        ("_VKTjIIQTEeyxv9w6U6_UQg", 26),
        ("_dr0dQLB8EeqLtsTDMgEHBw", 48),
        ("_gRrk4bB_EeqLtsTDMgEHBw", 11),
        ("__dKYsdZLEeqiU8uzTY0Puw", 14),
        ("_jGZ6geCUEeu-FsE4WGMncQ", 4),
        ("_NfMDoAATEeykFulkDFvkrg", 7),
        ("_I0wRYOQUEe2jpIKbeHI0hw", 14),
    ],
)
def test_iter_visible(model, diag_uid, num_nodes):
    diag = model.diagrams.by_uuid(diag_uid)
    diagram_elements = sum(
        1 for _ in aird.iter_visible(model._loader, diag._element)
    )

    assert diagram_elements == num_nodes



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Common classes used by all MelodyModel functions.

.. diagram:: [CDB] Common Types ORM
"""
from __future__ import annotations

import collections
import collections.abc as cabc
import typing as t

import capellambse

S = t.TypeVar("S", bound=t.Optional[str])
T = t.TypeVar("T", bound="ModelObject")
U = t.TypeVar("U")


XTYPE_ANCHORS = {
    "capellambse.model": "org.polarsys.capella.core.data.capellamodeller",
    "capellambse.model.crosslayer": "org.polarsys.capella.core.data",
    "capellambse.model.diagram": "viewpoint",
    "capellambse.model.layers": "org.polarsys.capella.core.data",
}
"""A mapping from anchor modules to Capella packages.

This dictionary maps Python modules and packages to the Capella packages
they represent. ``build_xtype`` and related functions/classes can then
use this information to automatically derive an ``xsi:type`` from any
class that is defined in such an anchor module (or a submodule of one).
"""
XTYPE_HANDLERS: dict[str | None, dict[str, type[t.Any]]] = (
    collections.defaultdict(dict)
)
r"""Defines a mapping between ``xsi:type``\ s and wrapper classes.

The first layer's keys can be either ``None`` or the ``xsi:type`` of the
architectural layer that the wrapper should be applied to. In the case
of ``None``, the wrapper will be applied to all layers. Note that
layer-specific wrappers have precedence over layer-agnostic ones.

These keys map to a further dictionary. This second layer maps from the
``xsi:type``\ (s) that each wrapper handles to the wrapper class.
"""


def build_xtype(class_: type[ModelObject]) -> str:
    anchor = package = ""
    for a, p in XTYPE_ANCHORS.items():
        if len(a) > len(anchor) and class_.__module__.startswith(a):
            anchor = a
            package = p

    if not anchor:
        raise TypeError(f"Module is not an xtype anchor: {class_.__module__}")

    module = class_.__module__[len(anchor) :]
    clsname = class_.__name__
    return f"{package}{module}:{clsname}"


def find_wrapper(typehint: str) -> tuple[type[ModelObject], ...]:
    """Find the possible wrapper classes for the hinted type.

    The typehint is either a single class name, or a namespace prefix
    and class name separated by ``:``. This function searches for all
    known wrapper classes that match the given namespace prefix (if any)
    and which have the given name, and returns them as a tuple. If no
    matching wrapper classes are found, an empty tuple is returned.
    """
    return tuple(
        v
        for i in XTYPE_HANDLERS.values()
        for k, v in i.items()
        if k.endswith(f":{typehint}") or k == typehint
    )


def enumliteral(
    generic_element: GenericElement, attr: str, default: str = "NOT_SET"
) -> AttributeProperty | str:
    uuid = generic_element._element.attrib.get(attr)
    if uuid is None:
        return default

    return generic_element.from_model(
        generic_element._model, generic_element._model._loader[uuid]
    ).name


def set_accessor(
    cls: type[ModelObject], attr: str, accessor: Accessor
) -> None:
    setattr(cls, attr, accessor)
    accessor.__set_name__(cls, attr)


def set_self_references(*args: tuple[type[ModelObject], str]) -> None:
    for cls, attr in args:
        set_accessor(cls, attr, DirectProxyAccessor(cls, aslist=ElementList))


def xtype_handler(
    arch: str | None = None, /, *xtypes: str
) -> cabc.Callable[[type[T]], type[T]]:
    """Register a class as handler for a specific ``xsi:type``.

    ``arch`` is the ``xsi:type`` of the desired architecture. It must
    always be a simple string or None. In the latter case the definition
    applies to all elements regardless of their architectural layer.
    Architecture-specific definitions will always win over
    architecture-independent ones.

    Each string given in ``xtypes`` notes an ``xsi:type`` of elements
    that this class handles. It is possible to specify multiple values,
    in which case the class will be registered for each ``xsi:type``
    under the architectural layer given in ``arch``.

    Handler classes' ``__init__`` methods must accept two positional
    arguments. The first argument is the
    :class:`~capellambse.model.MelodyModel` instance which loaded the
    corresponding model, and the second one is
    the LXML element that needs to be handled.

    Example::

        >>> @xtype_handler('arch:xtype', 'xtype:1', 'xtype:2')
        ... class Test:
        ...     _xmltag = "ownedTests"
        ...     def from_model(self, model, element, /):
        ...         ...  # Instantiate from model XML element
    """
    if arch is not None and not isinstance(arch, str):  # pragma: no cover
        raise TypeError(
            f"'arch' must be a str or None, not {type(arch).__name__}"
        )

    # Compile a list of all xtype strings
    xtype_strs = []
    for xtype in xtypes:
        if isinstance(xtype, str):
            xtype_strs.append(xtype)
        else:  # pragma: no cover
            raise ValueError(
                f"All `xtype`s must be str, not {type(xtype).__name__!r}"
            )

    def register_xtype_handler(cls: type[T]) -> type[T]:
        # Avoid double registration when executing an extension as module
        if cls.__module__ == "__main__":
            return cls

        if not xtype_strs:
            xtype_strs.append(build_xtype(cls))

        for xtype in xtype_strs:
            if xtype in XTYPE_HANDLERS[arch]:  # pragma: no cover
                raise LookupError(f"Duplicate xsi:type {xtype} in {arch}")
            XTYPE_HANDLERS[arch][xtype] = cls
        return cls

    return register_xtype_handler


from .accessors import *
from .accessors import _NewObject as new_object
from .element import *
from .properties import *

set_accessor(GenericElement, "parent", ParentAccessor(GenericElement))



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
import io
import pathlib
import zipfile

import pytest

from capellambse.filehandler import zip as zipfh

PPP = pathlib.PurePosixPath


@pytest.mark.parametrize("proto", ["file:", "zip:"])
def test_zipfilehandler_can_read_local_files(tmp_path, proto):
    with zipfile.ZipFile(tmp_path / "test.zip", "w") as z:
        z.writestr("test.txt", b"Hello, World!")
    zip_path = tmp_path.joinpath("test.zip").as_uri().replace("file:", proto)

    fh = zipfh.ZipFileHandler(zip_path)

    assert fh.read_file("test.txt") == b"Hello, World!"


def test_zipfilehandler_uses_zip_root_directory_as_subdir(tmp_path):
    with zipfile.ZipFile(tmp_path / "test.zip", "w") as z:
        z.writestr("mydir/test.txt", b"Hello, World!")
    zip_path = tmp_path.joinpath("test.zip").as_uri()

    fh = zipfh.ZipFileHandler(zip_path, subdir=None)

    assert fh.subdir == PPP("mydir")
    assert fh.read_file("test.txt") == b"Hello, World!"


def test_zipfilehandler_doesnt_use_zip_root_directory_if_explicitly_overridden(
    tmp_path,
):
    with zipfile.ZipFile(tmp_path / "test.zip", "w") as z:
        z.writestr("mydir/test.txt", b"Hello, World!")
    zip_path = tmp_path.joinpath("test.zip").as_uri()

    fh = zipfh.ZipFileHandler(zip_path, subdir=".")

    assert fh.subdir == PPP(".")
    assert fh.read_file("mydir/test.txt") == b"Hello, World!"


@pytest.mark.parametrize(
    "url",
    [
        "http://example.com/test.zip",
        "http://example.com/!test.zip",
        "http://example.com/%s!test.zip",
        "http://example.com/%s?v=1!test.zip",
    ],
)
def test_zipfilehandler_can_load_files_from_other_handlers(requests_mock, url):
    zipio = io.BytesIO()
    with zipfile.ZipFile(zipio, "w") as z:
        z.writestr("test.txt", b"Hello, World!")
    zipdata = zipio.getvalue()
    requests_mock.get("http://example.com/test.zip", content=zipdata)
    requests_mock.get("http://example.com/test.zip?v=1", content=zipdata)

    fh = zipfh.ZipFileHandler(url)

    assert fh.read_file("test.txt") == b"Hello, World!"



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ... import common as c
from ... import modeltypes
from . import datavalue


class DataType(c.GenericElement):
    _xmltag = "ownedDataTypes"

    is_discrete = c.BooleanAttributeProperty(
        "discrete",
        __doc__=(
            "Specifies whether or not this data type characterizes a discrete"
            " value (versus continuous value)"
        ),
    )
    min_inclusive = c.BooleanAttributeProperty("minInclusive")
    max_inclusive = c.BooleanAttributeProperty("maxInclusive")
    pattern = c.AttributeProperty(
        "pattern",
        default="",
        __doc__=(
            "Textual specification of a constraint associated to this data"
            " type"
        ),
    )
    visibility = c.EnumAttributeProperty(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )


@c.xtype_handler(None)
class BooleanType(DataType):
    literals = c.DirectProxyAccessor(
        datavalue.LiteralBooleanValue,
        aslist=c.ElementList,
        list_extra_args={"fixed_length": 2},
    )
    default = c.RoleTagAccessor("ownedDefaultValue")


@c.xtype_handler(None)
class Enumeration(DataType):
    """An Enumeration."""

    owned_literals = c.DirectProxyAccessor(
        datavalue.EnumerationLiteral, aslist=c.ElementList
    )

    sub: c.Accessor
    super: c.Accessor[Enumeration]

    @property
    def literals(self) -> c.ElementList[datavalue.EnumerationLiteral]:
        """Return all owned and inherited literals."""
        return (
            self.owned_literals + self.super.literals
            if isinstance(self.super, Enumeration)
            else self.owned_literals
        )


@c.xtype_handler(None)
class StringType(DataType):
    default_value = c.RoleTagAccessor("ownedDefaultValue")
    null_value = c.RoleTagAccessor("ownedNullValue")
    min_length = c.RoleTagAccessor("ownedMinLength")
    max_length = c.RoleTagAccessor("ownedMaxLength")


@c.xtype_handler(None)
class NumericType(DataType):
    kind = c.EnumAttributeProperty(
        "kind", modeltypes.NumericTypeKind, default="INTEGER"
    )
    default_value = c.RoleTagAccessor("ownedDefaultValue")
    null_value = c.RoleTagAccessor("ownedNullValue")
    min_value = c.RoleTagAccessor("ownedMinValue")
    max_value = c.RoleTagAccessor("ownedMaxValue")


@c.xtype_handler(None)
class PhysicalQuantity(NumericType):
    unit = c.RoleTagAccessor("ownedUnit")



#!/usr/bin/env python
# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import logging
import re
from typing import NamedTuple

import click
from lxml import etree

import capellambse
from capellambse import helpers


@click.command()
@click.argument(
    "metamodel",
    default="git+https://github.com/eclipse-capella/capella.git",
)
def main(metamodel: str):
    """Import all EEnum definitions from the given METAMODEL."""
    logging.basicConfig(level="INFO")
    handler = capellambse.get_filehandler(metamodel)

    print("import enum as _enum")
    print("class _StringyEnumMixin: ...")

    classes = {}
    for path in handler.rootdir.rglob("*.ecore"):
        with path.open("rb") as file:
            tree = etree.parse(file)
        for elem in tree.getroot().iter():
            if (
                elem.tag != "eClassifiers"
                or helpers.xtype_of(elem) != "ecore:EEnum"
            ):
                continue

            clsname = elem.attrib["name"]
            if clsname in classes:
                continue

            literals = [
                Literal(l.attrib["name"], l.get("value", "0"), getdoc(l))
                for l in elem.iterchildren("eLiterals")
            ]
            classes[clsname] = Class(clsname, literals, getdoc(elem))

    for i, cls in enumerate(sorted(classes.values())):
        if i:
            print("\n\n", end="")

        print("@_enum.unique")
        print(f"class {cls.name}(_StringyEnumMixin, _enum.Enum):")
        if cls.doc:
            print(f'    """{cls.doc}"""\n')
        for l in cls.literals:
            print(f"    {fixname(l.name)} = {l.name!r}")
            if l.doc:
                print(f'    """{l.doc}"""')


def getdoc(elem: etree._Element) -> str:
    try:
        annotation = next(
            i
            for i in elem.iterchildren("eAnnotations")
            if i.get("source")
            == "http://www.polarsys.org/kitalpha/ecore/documentation"
        )
    except StopIteration:
        return ""

    detail = annotation.xpath(".//details[@key='description']/@value")
    if not detail:
        return ""
    return detail[0].replace("\r", "")


def fixname(name: str) -> str:
    return re.sub(
        r"[a-z]+",
        lambda m: m.group(0).upper() + "_" * (m.span(0)[1] != len(name)),
        name,
    )


class Class(NamedTuple):
    name: str
    literals: list[Literal]
    doc: str


class Literal(NamedTuple):
    name: str
    value: str
    doc: str


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright DB InfraGO AG
# SPDX-License-Identifier: Apache-2.0
"""Global filter implementations."""
from __future__ import annotations

import lxml.etree

import capellambse.loader
from capellambse import diagram, helpers, model

from . import FilterArguments, composite, global_filter

XT_CEX_FEX_ALLOCATION = (
    "org.polarsys.capella.core.data.fa"
    ":ComponentExchangeFunctionalExchangeAllocation"
)
EXCHANGES_WITH_ROLES = frozenset(
    {"Aggregation", "Association", "Composition", "Generalization"}
)


@global_filter("hide.association.labels.filter")
def hide_association_labels(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Hide names of roles on an exchange."""
    del flt

    classes = EXCHANGES_WITH_ROLES
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge) or obj.styleclass not in classes:
            continue
        for label in obj.labels:
            label.hidden = True


@global_filter("hide.role.names.filter")
def hide_role_names(args: FilterArguments, flt: lxml.etree._Element) -> None:
    """Hide names of roles on an exchange."""
    del flt

    classes = EXCHANGES_WITH_ROLES
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge):
            continue

        if len(obj.labels) > 1 and obj.styleclass in classes:
            obj.labels = obj.labels[:1]


@global_filter("show.functional.exchanges.exchange.items.filter")
def show_name_and_exchangeitems_fex(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Change FEX labels to show Name and ExchangeItems."""
    del flt
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge):
            continue

        label_box = _get_primary_edge_label(obj, "FunctionalExchange")
        if label_box is None:
            continue

        sort_items = args.params.get("sorted_exchangedItems", False)
        ex_items = _stringify_exchange_items(
            obj, args.melodyloader, sort_items
        )
        if ex_items:
            label_box.label += " " + ex_items


@global_filter("show.exchange.items.filter")
def show_exchangeitems_fex(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Change FEX labels to only show ExchangeItems."""
    del flt
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge):
            continue

        label_box = _get_primary_edge_label(obj, "FunctionalExchange")
        if label_box is None:
            continue

        sort_items = args.params.get("sorted_exchangedItems", False)
        exchange_items_label = _stringify_exchange_items(
            obj, args.melodyloader, sort_items
        )
        if exchange_items_label:
            label_box.label = exchange_items_label


@global_filter("Show Exchange Items on Component Exchanges")
def show_exchangeitems_cex(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Change CEX labels to show ExchangeItems."""
    del flt
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge):
            continue
        label_box = _get_primary_edge_label(obj, "ComponentExchange")
        if label_box is None:
            continue

        assert obj.uuid is not None
        elm, items = _get_allocated_exchangeitem_names(
            obj.uuid,
            alloc_attr="convoyedInformations",
            melodyloader=args.melodyloader,
        )
        if elm is None:
            continue
        for fex in elm.iterchildren():
            if helpers.xtype_of(fex) != XT_CEX_FEX_ALLOCATION:
                continue
            items += _get_allocated_exchangeitem_names(
                fex.attrib["targetElement"],
                alloc_attr="exchangedItems",
                melodyloader=args.melodyloader,
            )[1]
        label_box.label = ", ".join(items)


@global_filter(
    "Show Exchange Items on Component Exchange without Functional Exchanges"
)
def show_exchangeitems_cex_no_fex(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Change CEX labels to show directly allocated ExchangeItems."""
    del flt
    for obj in args.target_diagram:
        if not isinstance(obj, diagram.Edge):
            continue
        label_box = _get_primary_edge_label(obj, "ComponentExchange")
        if label_box is None:
            continue

        assert obj.uuid is not None
        _, items = _get_allocated_exchangeitem_names(
            obj.uuid,
            alloc_attr="convoyedInformations",
            melodyloader=args.melodyloader,
        )
        label_box.label = ", ".join(items)


@global_filter("Hide Component Ports without Exchanges")
def hide_all_empty_ports(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Hide all ports that do not have edges connected."""
    del flt
    for dgobj in args.target_diagram:
        if isinstance(dgobj, diagram.Box) and dgobj.children:
            composite.hide_empty_ports(
                None, dgobj, classes=composite.PORT_CL_COMPONENT
            )


@global_filter("Hide Allocated Functional Exchanges")
def hide_alloc_func_exch(
    args: FilterArguments, flt: lxml.etree._Element
) -> None:
    """Hide functional exchanges that are allocated to a component exchange."""
    del flt

    component_exchanges = []
    for element in args.target_diagram:
        if not element.hidden and element.styleclass == "ComponentExchange":
            component_exchanges.append(element)

    for cex in component_exchanges:
        assert cex.uuid is not None
        # Find all allocated functional exchanges
        for fex in args.melodyloader[cex.uuid].iterchildren(
            "ownedComponentExchangeFunctionalExchangeAllocations"
        ):
            target = fex.attrib["targetElement"].rsplit("#", 1)[-1]
            try:
                args.target_diagram[target].hidden = True
            except KeyError:
                pass  # not in this diagram


def _stringify_exchange_items(
    obj: diagram.DiagramElement | model.common.GenericElement,
    melodyloader: capellambse.loader.MelodyLoader,
    sort_items: bool = False,
) -> str:
    assert obj.uuid is not None
    _, items = _get_allocated_exchangeitem_names(
        obj.uuid,
        alloc_attr="exchangedItems",
        melodyloader=melodyloader,
    )
    if items:
        if sort_items:
            items = sorted(items)

        return f"[{', '.join(items)}]"
    return ""


def _get_allocated_exchangeitem_names(
    *try_ids: str,
    alloc_attr: str,
    melodyloader: capellambse.loader.MelodyLoader,
) -> tuple[lxml.etree._Element | None, list[str]]:
    for obj_id in try_ids:
        try:
            elm = melodyloader[obj_id]
        except KeyError:
            pass
        else:
            break
    else:
        return (None, [])

    if elm.tag == "ownedDiagramElements":
        targetlink = next(elm.iterchildren("target"))
        elm = melodyloader[targetlink.attrib["href"]]

    names = []
    for elem in melodyloader.follow_links(
        elm, elm.get(alloc_attr, ""), ignore_broken=True
    ):
        name = elem.get("name")
        if name is not None:
            names.append(name)
    return (elm, names)


def _get_primary_edge_label(
    obj: diagram.Edge, edge_class: str
) -> diagram.Box | None:
    if obj.styleclass != edge_class:
        return None
    if not obj.labels:
        return None
    return obj.labels[0]
